@@ -2,15 +2,18 @@ import argparse
 import json
 import logging
 import os
+import shutil
+import sys
 import traceback
 from multiprocessing import cpu_count
 from pathlib import Path
 from typing import Dict, List, Optional, Sequence, Tuple, Union
 
 import librflxlang
-from pkg_resources import get_distribution
+from pkg_resources import get_distribution, resource_filename
 
 from rflx import __version__
+from rflx.converter import iana
 from rflx.error import ERROR_CONFIG, FatalError, RecordFluxError, Severity, Subsystem, fail
 from rflx.generator import Debug, Generator
 from rflx.graph import create_message_graph, create_session_graph, write_graph
@@ -51,6 +54,11 @@ def main(argv: List[str]) -> Union[int, str]:  # pylint: disable = too-many-stat
         metavar=("NUM"),
         help="parallelize proofs among NUM workers (default: %(default)d)",
     )
+    parser.add_argument(
+        "--unsafe",
+        action="store_true",
+        help="allow unsafe options (WARNING: may lead to erronous behavior)",
+    )
 
     subparsers = parser.add_subparsers(dest="subcommand")
 
@@ -190,6 +198,41 @@ def main(argv: List[str]) -> Union[int, str]:  # pylint: disable = too-many-stat
     )
     parser_validate.set_defaults(func=validate)
 
+    parser_setup = subparsers.add_parser(
+        "setup_ide", help="set up RecordFlux IDE integration (GNAT Studio)"
+    )
+    parser_setup.add_argument(
+        "--gnat-studio-dir",
+        dest="gnat_studio_dir",
+        type=Path,
+        help="path to the GNAT Studio settings directory (default: %(default)s)",
+        default=Path.home() / ".gnatstudio",
+    )
+    parser_setup.set_defaults(func=setup)
+
+    parser_convert = subparsers.add_parser(
+        "convert", help="convert foreign specifications into RecordFlux specifications"
+    )
+
+    convert_subparsers = parser_convert.add_subparsers(dest="subcommand")
+
+    parser_iana = convert_subparsers.add_parser(
+        "iana", help="convert IANA registry into RecordFlux specifications"
+    )
+
+    parser_iana.add_argument(
+        "-a",
+        "--always-valid",
+        help="add an Always_Valid aspect to each type that does not "
+        "explicitly cover all possible values for the type's length",
+        action="store_true",
+    )
+    parser_iana.add_argument(
+        "-d", dest="output_directory", type=Path, default=".", help="output directory"
+    )
+    parser_iana.add_argument("file", type=argparse.FileType("r"), nargs="?", default=sys.stdin)
+    parser_iana.set_defaults(func=convert_iana)
+
     args = parser.parse_args(argv[1:])
 
     if args.version:
@@ -203,6 +246,10 @@ def main(argv: List[str]) -> Union[int, str]:  # pylint: disable = too-many-stat
     if args.quiet:
         logging.disable(logging.CRITICAL)
 
+    if not args.unsafe:
+        if args.no_verification:
+            return 'cli: error: unsafe option "--no-verification" given without "--unsafe"'
+
     ERROR_CONFIG.fail_after_value = args.max_errors
 
     try:
@@ -210,6 +257,13 @@ def main(argv: List[str]) -> Union[int, str]:  # pylint: disable = too-many-stat
     except RecordFluxError as e:
         return f"{e}"
     except Exception:  # pylint: disable = broad-except
+        if args.unsafe:
+            return f"""
+----------------------------------------------------------------------------
+EXCEPTION IN UNSAFE MODE, PLEASE RERUN WITHOUT UNSAFE OPTIONS
+----------------------------------------------------------------------------
+{traceback.format_exc()}
+----------------------------------------------------------------------------"""
         return f"""
 ------------------------------ RecordFlux Bug ------------------------------
 {version()}
@@ -383,3 +437,23 @@ def validate(args: argparse.Namespace) -> None:
         fatal_error = FatalError()
         fatal_error.extend(e)
         raise fatal_error from e
+
+
+def setup(args: argparse.Namespace) -> None:
+    gnatstudio_dir = resource_filename("ide", "gnatstudio/")
+    plugins_dir = args.gnat_studio_dir / "plug-ins"
+    if not plugins_dir.exists():
+        plugins_dir.mkdir(parents=True, exist_ok=True)
+    print(f'Installing RecordFlux plugin into "{plugins_dir}"')
+    shutil.copy(Path(gnatstudio_dir) / "recordflux.py", plugins_dir)
+
+
+def convert_iana(args: argparse.Namespace) -> None:
+    xml_str = args.file.read()
+    iana.convert(
+        xml_str,
+        args.file,
+        args.always_valid,
+        args.output_directory,
+        os.environ.get("RFLX_REPRODUCIBLE") is not None,
+    )
