@@ -1,8 +1,11 @@
 import re
 import textwrap
 from abc import ABC
+from pathlib import Path
 from typing import Iterable, Iterator, Sequence, TypeVar
 
+STDIN = Path("<stdin>")
+
 
 class Base(ABC):
     def __eq__(self, other: object) -> bool:
