@@ -58,7 +58,7 @@ TYPES_LOW_ORDER_FIRST = TYPES * "Low_Order_First"
 UNREACHABLE = ada.Call(TYPES * "Unreachable")
 
 CONFIGURATION_PRAGMAS = [
-    ada.Pragma("Style_Checks", [ada.String("N3aAbcdefhiIklnOprStux")]),
+    ada.Pragma("Style_Checks", [ada.String("N3aAbCdefhiIklnOprStux")]),
     # https://github.com/Componolit/RecordFlux/issues/508
     ada.Pragma("Warnings", [ada.Variable("Off"), ada.String("redundant conversion")]),
 ]
