@@ -14,6 +14,7 @@ from .message import (  # noqa: F401
 )
 from .model import Model as Model  # noqa: F401
 from .session import (  # noqa: F401
+    FINAL_STATE as FINAL_STATE,
     AbstractSession as AbstractSession,
     Session as Session,
     State as State,
@@ -31,7 +32,6 @@ from .type_ import (  # noqa: F401
     Integer as Integer,
     ModularInteger as ModularInteger,
     Opaque as Opaque,
-    Private as Private,
     RangeInteger as RangeInteger,
     Scalar as Scalar,
     Sequence as Sequence,
